/*
 asm6502.c
  symbols like '+' (post-label), .shl (<<), .shr (>>),
   and the .lsb and .msb suffixes are not supported.

  Revised: 2013.05.13

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 TODO
	CHR banks
	mem
	index (mapper)
	segment (ignore)
	byte (explicit header)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "types.h"
#include "../interp/interpreter.h"
#include "../interp/ftrig.h"

#include "asm6502.h"

/* ---- macros */

#define IS_EOL(c)			((uint8)((c)-10)<=13-10)
#define SKIP_PRELAB(s,e)		do{while(s<e&&*s=='-') ++s;}while(0)

/* ---- code */

static const char *IDBreak(const char *str,
	const char *e)
{
	const char	*s = str;

	if (s>=e)
		return s;
	if (*s=='@') /* one starting 'at' only */
		if (++s>=e)
			return str;
	do
		switch(*s)
		{
		case '.': /* only one consecutive '.' */
			//if (s>str && s[-1]=='.')
			//	return s;
		case '_':
		CASE_NUMERIC
		CASE_ALPHABETIC
		next:	++s;
			break;

		default:
#ifdef CHR_NB
			if (*s==CHR_NB)
				goto next;
#endif
			return s;
		}
	while(s<e);

	return s;
}

long X6502_ParseName(const char *str,
	const char *e,
	long *z,
	char **b)
{
	const char	*s;
	long		r;

	if (!str || str>=e)
		return _e_general_error;
	if (*str=='\047' || *str=='"')
	{
		r =-1u>>1;
		s = parse_textloop(str,e,&r,NULL,b,0,NULL);
		if (s>=e)
			return _e_open_string;
		if (*s++!=*str) /* equal close quote or point of error */
			return _e_general_error;
		if (r>=0)
		{
			char	*p;

			if ((p=realloc(*b,r+1))==NULL)
				return _e_memory_error;
			(*b=p)[r] = 0;
			if (z)
				*z = r;
			return s-str;
		}
	} else /* use ID break - can expect an extension */
		if ((r=IDBreak(str,e)-str)>0)
		{
			*b = malloc(r+1);
			memcpy(*b,str,r);
			(*b)[r] = 0;
		}
	return r;
}

static const char *parse_value(const char *s,
	const char *e,
	evalvalue *i,
	void *u)
{
	uint8		b[256], *d;
	const char	*n;
	int32		x;

	if (s>=e)
		return NULL;
	if (*s=='*')
	{ /* current location */
		if (!u)
			return NULL;
		if (i)
			*i = ((_asm6502frame*)u)->org;
		return s+1;
	} else if (*s=='%')
	{ /* binary */
		uint32	v = 0;
		uint8	c;

		if (++s>=e)
			return NULL;
		do
			if ((c=*s-'0')>1)
			{ /* can have no other following alphanumerics */
				if (c<=9 || (c+='0'-'A')<='Z'-'A'
						|| (c+='A'-'a')<='z'-'a')
					s = NULL;
				break;
			} else
				v+= v, v+= c;
		while(++s<e);
		if (i)
			*i = v;
		return s;
	}
	n = IDBreak(s,e);
	if (s==n && *s=='-')
		SKIP_PRELAB(n,e);
	if ((x=X6502_AsmGetID(s,n,b,u))>0)
		d = ((uint8*)((_asm6502frame*)u)->lab)+x+(n-s);
	else if (!(((_asm6502frame*)u)->flg&_Prescan)
			|| (x=X6502_AsmAddID(s,n,b,u,_Unestablished))<=0)
		return NULL;
	else
	{
		d = ((uint8*)((_asm6502frame*)u)->lab)+x;
		d[2] = ((_asm6502frame*)u)->org>>8; /* page copy */
	}
	if (*d==_Unestablished)
		((_asm6502frame*)u)->flg|= _UnestID;
	if (i)
		*i = d[1]|((d[2]|((d[3]|(d[4]<<8))<<8))<<8);
	return n;
}

#if 0
static char	modes[][3] =
{
	"IM", /* 1-immediate		- #$xx */
	"RL", /* 1-relative		- *+$xx */
	"ZP", /* 1-zero page		- $xx */
	"ZX", /* 1-zero page, X index	- $xx,X */
	"ZY", /* 1-zero page, Y index	- $xx,Y */
	"IX", /* 1-indirect, X index	- ($xx,X) */
	"IY", /* 1-indirect, Y index	- ($xx),Y */
	"IN", /* 1-indirect		- ($xx) */
	"IB", /* 2-absolute indirect	- ($xxxx) */
	"AB", /* 2-absolute		- $xxxx */
	"AX", /* 2-absolute, X index	- $xxxx,X */
	"AY", /* 2-absolute, Y index	- $xxxx,Y */
	"IA", /* 0-implied A to self */
	"IP"
};
#endif

/* returns the total parsed length, and larger operand mode in `m` */
long X6502_ParseOperands(const char *str,
	const char *e,
	evalvalue *v,
	_asm6502frame *f,
	int *m)
{
	const char	*s = str, *n;
	long		r;
	int		t;
	char		b = 0, c;

	if (s>=e || (c=*s)==';' || IS_EOL(c))
		{*m = OPM_IP; return 0;}
	if ((n=memchr(s,';',e-s))==NULL)
		n = e;
	else
		e = n;
	if (c=='#')
		*m = OPM_IM, ++s;
	else if (c=='(')
	{ /* indirect - must scan for each evaluating bounds */
		b = ')';
	ib:	*m = OPM_IB, n = ++s, r = 1;
		for(; n<e && (c=*n)!=','; ++n)
			if (c=='(')
				++r;
			else if ((c==')' || c==']') && !--r)
				break;
	} else if (c=='[')
		{b = ']'; goto ib;}
	else
		*m = OPM_AB;
	if ((r=evaluate_line(s,n,v,&t,parse_value,NULL,f,NULL))==0 || t)
	{ /* for pre-labels - you may get awat with `[-]` */
		if ((n=parse_value(s,n,v,f))==NULL)
			return t<0 ? t:_e_general_error;
		else
			r = n-s;
	}
	s = parse_skipwsp(s+r,e);
	r = e-str;
	if (s>=e /*|| *s==';'*/ || IS_EOL(*s))
		return b ? _e_general_error:r;
	if (*s==')' || *s==']')
	{ /* bracket/parenthesis must match, IB or IY, resulting */
		if (!b || *s!=b || *m!=OPM_IB)
			return _e_general_error;
		t = 1, s = parse_skipwsp(s+1,e);
	}
	if (*s==',')
	{
		s = parse_skipwsp(s+1,e);
		if (s>=e) /* only a comma? */
			return _e_general_error;
		if ((c=toupper(*s))=='X' && !t)
			if (*m==OPM_AB)
				*m = OPM_AX;
			else if (*m==OPM_IB)
				*m = OPM_IX;
			else
				return _e_general_error;//_e_op_error;
		else if (c=='Y')
		{ /* must be closed first */
			if (*m==OPM_AB)
				*m = OPM_AY;
			else if (*m==OPM_IB && t)
				*m = OPM_IY;
			else
				return _e_general_error;//_e_op_error;
		} else
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
		/* need indirect closing?  IY doesn't. */
		if (s>=e || IS_EOL(*s) /*|| *s==';'*/)
			return *m==OPM_IX ? _e_general_error:r;
		if (*s!=b) /* IX does */
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
	} else if (b) /* need indirect closing */
		return _e_general_error;
	if (s<e && !IS_EOL(*s) /*&& *s!=';'*/)
		return _e_general_error;
	return r;
}

/* ---- keywords */

static const char	*asm6502_keyw[] =
{
	"INCBIN",
	"INCLUDE",
	"BANK",
	"BLOCK",
	"BSS",
	"CODE",
	"DB",
	"DCB",
	"DC.B",
	"DEFAULT",
	"DEFCHR",
	"DW",
	"DCW",
	"DC.W",
	"END",
	"EQU",
	"FILL",
	"MODULE",
	"OFF",
	"ON",
	"OPT",
	"ORG",
	"P02",
	"PAD",
	"PROCESSOR",
	"SUBROUTINE",
	"ZEROPAGE",
	"ZP"
};

#define NUMBER_OF_KEYS		(OPN_COUNT+GET_ARRAY_COUNT(asm6502_keyw))

int X6502_ParseKey(const char *s,
	const char *e,
	uint32 max)
{
	char		n[10];
	int		i, z = e-s;
	uint8		m = 0;

	if (!s || s>=e)
		return _e_general_error;
	if (max>NUMBER_OF_KEYS)
		max = NUMBER_OF_KEYS;
	if (*s=='.')
		++s, m = 1;
	if ((z=e-s)>9)
		goto nope;
	i = 0;
	do
		n[i] = toupper(*s++);
	while(++i,s<e);

	i = 0;
	if (z==3 && !m)
	{
		const char	*t = *x6502_opnametable;

		do
			if (*t==*n && t[1]==n[1] && t[2]==n[2])
				return i;
		while(t+=3,++i<OPN_COUNT);
	} else
		i = OPN_COUNT;
	if (max
#ifdef KEYW_REQUIRES_DOT
			&& m
#endif
			)
	{ /* keyword (e.g., .equ) */
		const char	**t = asm6502_keyw, c = *n;

		--z;
		do
			if (**t==c && !memcmp(*t+1,n+1,z))
				return i;
		while(++t,++i<max);
	}
nope:	return _e_no_identifier;
}

/* returns ID pstr offset, ~end or zero on error */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	uint8 *n,
	const _asm6502frame *f)
{
	const uint8	*l;
	uint8		h, z;

	if (!f || !n || !s || e<=s)
		return 0;
	*n = toupper(*s++)&0x3F, z = h = 0;
	while(s<e)
		if (!++z)
			return 0;
		else
			h+= n[z] = toupper(*s++);
	if ((h=h<<6|*n)==0) /* '@' is valid */
		h = 0x80; /* hash can't be null */
	*n++ = h;
	if ((l=f->lab)==0)
		return -1;
	/* hash and type match */
	for(; *l++; l+=*l,l+=1+1+4)
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return l-f->lab;
	return f->lab-l; /* becomes ~total */
}

/* returns ID offset, or <0 on error */
int X6502_AsmAddID(const char *s,
	const char *e,
	uint8 *n,
	_asm6502frame *f,
	uint8 t)
{
	
	uint8		*d;
	int32		x;

	if ((x=X6502_AsmGetID(s,e,n,f))>=0)
	{
		d = f->lab+x+(e-s);
		if (t && *d) /* fully established? */
			return -x;
		if (f->flg&_Prescan)
			goto end;
		goto typ; /* can only establish once */
	}
	/* (-x) end offset including null term., 1 length, 1 typ, 4 val */
	if ((d=realloc(f->lab,(e-s)+1+1+4-x))==0)
		return _e_memory_error;
	f->lab = d, d-= x;
	d[-1] = *n++; /* hash (includes first char) */
	*d++ = x = e-s-1; /* post-hash length */
	memcpy(d,n,x); /* post-hash name */
	d+= x;
	d[1] = d[2] = d[3] = d[4] = d[5] = 0; /* init and terminate */
typ:	*d = t;
end:	return d-f->lab;
}

/* ---- assembler */

static int getInstruction(int t)
{
	int		i = 0;

	do
		if ((x6502_opn_table[i]&(OPB_BAD-1))==t)
			return i;
	while(++i<256);

	return -1;
}

static long incbin(const char *str,
	const char *e,
	int (*writef)(uint16,
		uint8,
		uint8),
	_asm6502frame *f)
{
	FILE		*nf = NULL;
	char		*b = NULL;
	long		z;

	if ((z=X6502_ParseName(str,e,NULL,&b))>0)
		nf = fopen(b,"r");
	if (b)
		free(b);
	if (!nf)
		return z ? (z>0 ? _e_inc_open:z):_e_incomplete_error;
	if (!writef)
	{
		fseek(nf,0,SEEK_END);
		f->org+= ftell(nf);
	} else
	{
		int		c;

		while((c=fgetc(nf))>=0)
			writef(f->org,f->bnk,c), ++f->org;
	}
	return z;
}

#define _LineFlags			(_IntoLine|_UnestID)

int X6502_Assemble(const char *s,
	int (*writef)(uint16,
		uint8,
		uint8),
	_asm6502frame *f)
{
	uint8		b[256];
	_asm6502frame	af;
	const char	*e;
	int		r, m, wrn = 0;
	uint8		c = 0;

	if (!s)
		return -1;
	r = 0;
	if ((e=strlen(s)+s)==s)
		return 0; /* empty */
	if (f)
		f->flg&=~_UnestID;
	else
	{ /* clear */
		f = &af;
		f->org = 0, f->flg = 0;
		f->lab = 0;
	}

	while(s<e)
	{
		evalvalue	v;
		const char	*n;
		long		l;
		int		x;

		if (IS_EOL(*s)) /* new line */
			{f->flg&=~_LineFlags, ++s; continue;}
		if (c)
			{++s; continue;}
		if (*(n=parse_skipwsp(s,e))==';')
			{c = 1, s = n+1; continue;}
		if (n!=s)
			f->flg|= _IntoLine;
		m =-1, v = f->org, n = IDBreak(s,e);
		if ((x=X6502_ParseKey(s,n,-1))>=0)
		{ /* keyword letters starting where a label would */
		  /* some assembler(s) allow this */
			//if (!(f->flg&_IntoLine) && *s!='.')
			//	goto labe;
			if (x==_default)
				s = parse_skipwsp(n,e), n = IDBreak(s,e);
			else
				goto gotn;
		}
		if (*n==':' || !(f->flg&_IntoLine))
		{ /* identifier: equate/label */
			long		z;
			uint8		t = _Unestablished;

			if ((z=n-s)==0) /* '-' mark (pre-)labels */
				if (*n==':')
					s = "-", z = 1; /* `s` not reused */
				else
				{
					SKIP_PRELAB(n,e);
					z = n-s;
				}
			else if (x!=_default)
				t = _Label;
			if ((m=X6502_AsmAddID(s,s+z,b,f,t))<=0)
				{r = m ? _e_identifier_conflict:0; break;}
			if (*n==':')
				++n;
		}
		if (*(n=parse_skipwsp(n,e))==';')
			{c = 1, s = n+1; continue;}
		s = n;
		if ((n=IDBreak(n,e))<=s)
		{
			if (*s=='=') /* '=' equate */
			{
				++n;
				if (x!=_default)
					x = _equ;
				goto gotn;
			}
			if (m>=0)
				goto lset;
			r = _e_general_error;
			break;
		} else if (x==_default) /* must follow '=' */
			{r = _e_general_error; break;}
		/* macro or instruction */
		if ((x=X6502_ParseKey(s,n,-1))<0)
			{r = x==_e_no_identifier ? _e_macro_error:x; break;}
		/* keyword, possible equate, etc. */
	gotn:	r = 0, s = parse_skipwsp(n,e);
		/* now at position of the parameter(s) */
		switch(x)
		{ /* the `r` end-value when using lset displaces org */
		case _bank:
			s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			f->bnk = v;
			break;
		case _block: /* asign block, skip */
			s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			r = v, v = f->org, x = IMAX(x);
			goto lset;
		case _bss:
			f->bnk = 0, f->org&= 0x1FFF;
			break;
		case _code:
			if (f->org<0x8000)
				f->org = 0x8000, f->bnk = 0;
			break;
		case _db:
		case _dcb:
		case _dc_b:
			x = 0;
			do
			{
				char		*t;

				if (x && *s++!=',')
					{r = _e_general_error; goto end;}
				l = 1, s = parse_skipwsp(s,e);
				if (*s=='\047' || *s=='"')
				{ /* TODO: use struct func instead of X6502_ParseName */
					t = NULL;
					if ((r=X6502_ParseName(s,e,&l,&t))==0)
						r = _e_general_error;
					else if (r>0)
						s+= r;
				} else
				{
					s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
					t = (char*)b, *b = v;
				}
				if ((unsigned)r>=_e_return)
				{
					if (t!=(char*)b)
						free(t);
					goto end;
				}
				if (t)
				{
					if (!writef)
						f->org+= l;
					else
					{
						r = 0;
						do
							writef(f->org,f->bnk,t[r]);
						while(++f->org,++r,--l);
					}
					if (t!=(char*)b)
						free(t);
					r = 0;
				}
			} while(++x,(s=parse_skipwsp(s,e))<e
					&& !IS_EOL(*s) && *s!=';');
			x = _db;
			break;
		case _defchr: /* untested */
			l = 0;
			do
			{ /* supposed to pick up 8 hex codes, proc. 2 bits of ea. nybble */
				uint16	dp;

				if (l && ((s=parse_skipwsp(s,e))>=e || *s++!=','))
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
				dp = 0, x = 8;
				do
					dp|= (v>>28-8)&0x100, dp|= (v>>28+1)&1;
				while(dp+=dp,v<<=4,--x);
				b[l] = dp>>8, b[l+8] = dp;
			} while(++l<8);
			if (!writef)
				f->org+= 16;
			else
			{
				x = 0;
				do
					writef(f->org,f->bnk,b[x]);
				while(++f->org,++x<16);
			}
			x = _defchr;
			break;
		case _dw:
		case _dcw:
		case _dc_w:
			l = 0;
			do
			{
				if (l && *s++!=',')
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
				if (!writef)
					f->org+= 2;
				else
				{
					writef(f->org,f->bnk,v), ++f->org;
					writef(f->org,f->bnk,v>>8), ++f->org;
				}
			} while(++l,(s=parse_skipwsp(s,e))<e
					&& !IS_EOL(*s) && *s!=';');
			break;
		case _end:
			f->flg|= _EndKeyw;
			break;
		case _default:
		case _equ: /* set label to explicit value */
			s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (m<0)
				{r = _e_no_identifier; goto end;}
			if (x!=_default && f->lab)
				f->lab[m] = _Equate;
			x = IMAX(x), r = 0;
			goto lset;
		case _pad: /* pad to this next address */
		case _fill: /* fill bytes of length, optional value */
			s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (x==_fill)
				x = v;
			else if ((x=v-f->org)<0)
				x = 0;
			v = 0;
			if (*(s=parse_skipwsp(s,e))==',')
			{
				++s;
				s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
			}
			if (!writef || !x)
				f->org+= x;
			else
				do
					writef(f->org,f->bnk,v);
				while(++f->org,--x);
			x = IMAX(x), r = 0;
			goto lset;
		case _incbin:
			if ((r=incbin(s,e,writef,f))<=0)
				goto end;
			x = IMAX(x), r = 0;
			goto lset;
		//case _off: /* on/off should not be used as labels */
		//case _on:
		case _include: /* read this file in place of line */
			/* this should already have been processed;
			  we can assume a label preceeded, which is illegal here */
		//labe:
			r = _e_label_error;
			goto end;
		case _module:
			r = _e_handler_error;
			goto end;
		case _opt:
			r = X6502_ParseKey(s=parse_skipwsp(s,e),e,-1);
			if ((unsigned)(r-=_off)>1)
				{r = _e_on_off; goto end;}
			if (r)
				f->flg|= _Optimize;
			else
				f->flg&=~_Optimize;
			break;
		case _processor:
			if ((n=IDBreak(s,e))==s)
				{r = _e_general_error; goto end;}
			if (n-s!=4 || !memcmp(s,"6502",4))
				{r = _e_language_error; goto end;}
			break;
		case _subroutine:
			wrn =~_e_cannot_optimize;
			break;
		case _zp:
		case _zeropage:
			f->org = 0, f->bnk = 0;
			break;
		case _org: /* set position */
			s+= evaluate_line(s,e,&v,&r,parse_value,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			/* there is also an optional fill value... */
			f->org = v, r = 0, x = IMAX(x);

		default: /* label at org */
		lset:	if (m>=0)
			{
				uint8	*d;

				if ((d=f->lab)==NULL)
					{r = _e_read_error; break;}
				d+= m;
				d[1] = v, d[2] = v>>8, d[3] = v>>16, d[4] = v>>24;
			}
			f->org+= r;
		}
		s = parse_skipwsp(s,e);
		if ((unsigned)x>=OPN_COUNT)
		{ /* not encoding an instruction */
			if (s>=e || IS_EOL(*s))
				{f->flg&=~_LineFlags, ++s; continue;}
			if (*s==';')
				continue;
			r = _e_general_error;
			goto end;
		}
		r = X6502_ParseOperands(s,e,&v,f,&m);
		if ((unsigned)r>=_e_return)
			goto end;
		s+= r;

		/* optimize */
		if ((unsigned)m<OPM_IA && (unsigned)v<256
				&& f->flg&_Optimize && !(f->flg&_UnestID))
			if (m==OPM_AB)
				m = OPM_ZP; /* optimize size */
			else if (m==OPM_IB)
				m = OPM_IN;
			else if (m==OPM_AX)
				m = OPM_ZX;
			else if (m==OPM_AY)
				m = OPM_ZY;
		if ((unsigned)(r=getInstruction(x|(m<<8)))>255)
			if (m==OPM_IP && (unsigned)(r=getInstruction(x|(OPM_IA<<8)))<256)
				m = OPM_IA;
			else if ((m==OPM_ZP || m==OPM_AB)
					&& (unsigned)(r=getInstruction(x|(OPM_RL<<8)))<256)
			{ /* instruction calls for a relative; convert */
				m = OPM_RL, v-= f->org, v-= 2;
				/* check range */
				if (v!=(int8)v && !(f->flg&_Prescan))
					{r = _e_range_error; break;}
			} else if (f->flg&_Optimize)
			{ /* last resort: if reduced, check for a larger, alt. mode */
				if (m==OPM_ZP)
					m = OPM_AB;
				else if (m==OPM_IN)
					m = OPM_IB;
				else if (m==OPM_ZX)
					m = OPM_AX;
				else if (m==OPM_ZY)
					m = OPM_AY;
				if ((unsigned)(r=getInstruction(x|(m<<8)))>255)
					{r = _e_op_error; break;} /* no such instruction */
			}
		if (writef)
			writef(f->org,f->bnk,r);
		++f->org, r = 0;
		if (m>=OPM_IA)
			continue;
		if (writef)
			writef(f->org,f->bnk,v);
		++f->org;
		if ((unsigned)(m-OPM_IB)<=OPM_AY-OPM_IB)
		{ /* 2-byte mode range */
			if (writef)
				writef(f->org,f->bnk,v>>8);
			++f->org;
 			if ((unsigned)v>0xFFFF)
				wrn =~_e_range_error;
		} else if ((unsigned)v>0xFF && m!=OPM_RL)
			wrn =~_e_range_error;
	}
	f->flg&=~_IntoLine;
end:	if (f==&af && af.lab)
		free(af.lab);
	return r ? r:wrn;
}

#ifdef SAMPLE
 #pragma warn(disable:2154)

static long uppow2(long v)
{
	int		x = 1u<<(31-cntlz32(v));

	if (v!=x)
		x+= x;
	return x;
}

static uint8	*pbuf;
static uint32	pbmk;

int bwrite(uint16 o,
	uint8 b,
	uint8 v)
{
	int32		a;

	if (pbuf && (a=o-0x8000)>=0)
	{
		a|= (b<<14)&0x4000; /* 16 KB banks */
		pbuf[a&pbmk] = v;
		return 1;
	}
	return 0;
}

static const char	*err_strs[] =
{
	"syntax/param error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"bad or missing ID",
	"handler not supplied",
	"too many/few parentheses",
	"value out of range",
	"comment still open",
	"unterminated string",
	"ID already defined",
	"no such instruction mode",
	"invalid label asignment",
	"unknown/unsupported language",
	"unrecognized macro, keyword or instruction",
	"subroutines are not managed"
};

#define MAXIMUM_DEPTH		8

int _asm6502(FILE *f,
	const char *ifn,
	char **str,
	_asm6502frame *af,
	uint8 depth)
{
	uint32		z, l, x = 1;

	if (depth>=MAXIMUM_DEPTH)
		return _e_include_depth;
	while(z=~0,l=x,f && (*str=freadline(f,&z,*str,NULL,&x,'\\'))!=NULL)
	{
		const char	*e = *str+z;
		const char	*s = parse_skipwsp(*str,e);
		const char	*n = IDBreak(s,e);
		long		r;

		if (X6502_ParseKey(s,n,_include+1)==_include)
		{
			FILE	*nf = NULL;
			char	*b = NULL;

			if ((z=X6502_ParseName(parse_skipwsp(n,e),e,NULL,&b))>0)
				nf = fopen(b,"r");
			if (nf)
			{
				r = _asm6502(nf,b,str,af,depth+1);
				fclose(nf);
			} else
				r = z ? (z>0 ? _e_inc_open:z):_e_incomplete_error;
			if (b)
				free(b);
			if (r>0) /* error/warning message already handled? */
				return r;
		} else
			r = X6502_Assemble(s,af->flg&_Prescan ? NULL:bwrite,af);
		if (r)
		{
			const char	*errs = NULL, *errh;
			long		errc;

			errh = af->flg&_Prescan ? NULL:
"## >%s\n"
"## Warning - %s\n"
"File '%s' Line %u\n\n";
			if (r==_e_param_separation || r==_e_return)
				r = _e_general_error;
			if ((errc=r)<0)
				errc =~errc, errh =
"\n"
"###>%s\n"
"### %s\n"
"File '%s' Line %u\n\n";
			if (errc<GET_ARRAY_COUNT(err_strs))
				errs = err_strs[errc];
			else if (r<0)
				errs = r==_e_on_off ? "expected 'on' or 'off'"
					:r==_e_include_depth ? "exceeded maximum include depth"
					:r==_e_inc_open ? "could not open include file"
					:"unknown error";
			if (errh && errs)
				fprintf(stderr,errh,
					*str,errs,ifn,l);
			if (r<0)
				return 1;
		}
		if (af->flg&_EndKeyw)
			return 0;
	}
	return z ? _e_memory_error:0; /* length - realloc failed */
}

static int writeoutput(const char *n,
	long m)
{
	FILE		*f;
	int		r = 0;

	if ((f=fopen(n,"w"))==NULL)
	{
		fprintf(stderr,"## writeoutput: could not open '%s' for writing.\n",
			n);
		return 1;
	}
	if (fwrite(pbuf,1,m,f)!=m)
		r = _e_write_error;
	fclose(f);
	return r;
}

static int asm6502(const char *ifn,
	const char *ofn,
	long m)
{
	_asm6502frame	af;
	FILE		*f;
	char		*str;
	int		r;

	if (!memcmp(ifn,"dev::stdin",11))
		f = stdin;
	else
		f = fopen(ifn,"r");
	if (!f)
	{
		fprintf(stderr,"### could not open file '%s'\n",
			ifn);
		return 1;
	}
	if ((m=uppow2(m))<1024)
		m = 1024;
	pbmk = m-1;
	if ((pbuf=calloc(1,m))==NULL)
		return _e_memory_error;
	printf("# PRG size set to %u KB\n",
		m>>10);
	memset(&af,0,sizeof(_asm6502frame));
	str = malloc(0); /* needed for first-empty line(s) case */

	/* at code point, initially */
	af.org = 0x8000, af.flg = _Prescan;
	if ((r=_asm6502(f,ifn,&str,&af,0))==0)
		af.org = 0x8000, af.flg = 0, r = _asm6502(f,ifn,&str,&af,0);

	free(str);
	if (af.lab)
		free(af.lab);
	if (f && f!=stdin)
		fclose(f);
	if (!r)
		r = writeoutput(ofn,m);
	if (pbuf)
		{free(pbuf); pbuf = NULL;}
	return r;
}

int main(int argc,
	char **argv)
{
	const char	*ofn = "asm.out", *ifn = NULL, *s;
	char		**a;
	long		m = 1u<<15;
	int		r =-1;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (!--argc || (s=*++a)==NULL)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		goto disp;
	}
	do
		if (*s=='-')
		{
			if (!s[1])
				{s = "dev::stdin"; goto gotn;}
			else if (s[1]=='O')
				ofn = s+2, s = NULL;
			else if (s[1]=='P')
				m = atoi(s+2)<<10, s = NULL;
			else
			{
				fprintf(stderr,"### %s - unknown option '%s'.\n",
					*argv,s);
				return 1;
			}
		} else
		{
		gotn:	if (ifn)
			{
				fprintf(stderr,"### %s - one input at a time.\n",
					*argv);
				goto bail;
			}
			ifn = s;
		}
	while(--argc);

	if (!ifn)
	{
		fprintf(stderr,"### no file input\n",
			*argv);
	disp:	fputs(
"#\n"
"#  -            use stdin\n"
"#  -Z           set output size\n"
"#  -O           set output name\n"
"#\n",
			stderr);
		return 1;
	}
	r = asm6502(ifn,ofn,m);
bail:	if (r<0)
		fprintf(stderr,"### %s - fatal error (%d).\n",
			*argv,r), r = 1;
	return r;
}
#endif
